---- Generates tile geoemetry from Quadkey

CREATE OR REPLACE FUNCTION public.generate_tile_geometry_from_quadkey(quadkey TEXT)
RETURNS GEOMETRY -- Using Web Mercator, EPSG 3857
AS $$
    DECLARE
        earth_rad FLOAT; x FLOAT; y FLOAT; off_set FLOAT; q TEXT;
    BEGIN
        SELECT 6378137.0 INTO earth_rad; -- EPSG 3857 definition
        SELECT 0 INTO x;
        SELECT 0 INTO y;
        SELECT earth_rad * pi() INTO off_set; -- Tile centroid
        FOR i IN 1..LENGTH($1) LOOP
            q = SUBSTRING(quadkey, i, 1);
            off_set = off_set * .5;
            IF q = '0' THEN
                x = x - off_set; y = y + off_set;
            ELSIF q = '1' THEN
                x = x + off_set; y = y + off_set;
            ELSIF q = '2' THEN
                x = x - off_set; y = y - off_set;
            ELSIF q = '3' THEN
                x = x + off_set; y = y - off_set;
            ELSE RAISE EXCEPTION 'Failure: a Quadkey can only contain numbers from `0` to `3`.';
            END IF;
        END LOOP;
        RETURN ST_MakeEnvelope(x - off_set, y - off_set, x + off_set, y + off_set, 3857);
    END
$$
LANGUAGE plpgsql IMMUTABLE PARALLEL SAFE;

---- Generates vector tiles containing their own geometry

CREATE OR REPLACE FUNCTION public.generate_vector_tile_with_geometry(z INTEGER, x INTEGER, y INTEGER)
RETURNS BYTEA AS
$$
    DECLARE
        vector_tile BYTEA DEFAULT '';
    BEGIN
        WITH mvtdata AS (
            SELECT ST_MakeEnvelope(0, 0, 4096, 4096) AS geom,
            z as zoom
        )
        SELECT ST_AsMVT(mvtdata.*, 'tiles' 4096, 'geom') FROM mvtdata INTO vector_tile;
        RETURN vector_tile;
    END;
$$
LANGUAGE plpgsql IMMUTABLE PARALLEL SAFE;

---- Generates vector tiles containing geometries of tiles with a specified zoom level and allows for vizualization on lower zoom levels

CREATE OR REPLACE FUNCTION public.generate_vector_tile_containing_geometries_of_sub_tiles(z INTEGER, x INTEGER, y INTEGER, target_zoom_level INTEGER)
RETURNS BYTEA AS
$$
    DECLARE
        vector_tile BYTEA DEFAULT '';
        inside_tiles INTEGER; -- Number of tiles at zoom level `target_zoom_level` along one axis of a tile with zoom level `z`.
        inside_tile_pixels INTEGER;
    BEGIN
        IF z > target_zoom_level THEN
            RAISE NOTICE 'Vector tiles cannot be be created for zoom levels higher than the indicated in `target_zoom_level`: %.', target_zoom_level;
        ELSE
            inside_tiles = 2 ^ (target_zoom_level - z);
            inside_tile_pixels = 4096 / inside_tiles;
            WITH mvtdata AS (
                SELECT ST_MakeEnvelope(
                    inside_tile_pixels * tile_x, inside_tile_pixels * tile_y,
                    inside_tile_pixels * (tile_x + 1), inside_tile_pixels * (tile_y + 1)
                ) AS geom
                FROM
                    generate_series(0, inside_tiles) AS tile_x,
                    generate_series(0, inside_tiles) AS tile_y
            )
            SELECT ST_AsMVT(mvtdata.*, 'tiles', 4096, 'geom') FROM mvtdata INTO vector_tile;
        END IF;
        RETURN vector_tile;
    END;
$$
LANGUAGE plpgsql IMMUTABLE PARALLEL SAFE;
